/*
	* Copyright (c) 2021-11 by ELECTRONIC ASSEMBLY <technik@lcd-module.de>
	*
	* WiFi Access to ESP32. Wrapper for WiFi.h
	*
	* This file is free software; you can redistribute it and/or modify
	* it under the terms of either the GNU General Public License version 2
	* or the GNU Lesser General Public License version 2.1, both as
	* published by the Free Software Foundation.
*/

#include "../../debug.h"
#include <ESPmDNS.h>
#include <ETH.h>
#include <WiFi.h>

#include "wifi_ea.h"

#include "../SmallProtocol/uniTFT.h"

WiFiServer Server(SERVERPORT); //TCP server for direct connection LAN to uniTFT
WiFiClient WiFiRemoteClient;

wifi_ea WiFi_EA;

/*----------------------------
	Func: WiFiEvent
	Desc: Ethernet event structure. Wireless is off, if cable is connected to lan.
	Vars: ---
	Ret : ---
------------------------------*/
void WiFiEvent(WiFiEvent_t event)
{
  switch (event) {
    case ARDUINO_EVENT_ETH_START:
		#if (DBG_ON && DBG_WiFi)
			Serial.println("Ethernet: Started");
		#endif
		ETH.setHostname(ETHERNETHOSTNAME);
		break;
    case ARDUINO_EVENT_ETH_CONNECTED:
		#if (DBG_ON && DBG_WiFi)
			Serial.println("Ethernet: Connected");
		#endif
	  	if(WiFi.status() == WL_CONNECTED) //already connected to wifi
		{
			WiFi_EA.setWifi_reconnect_needed(false); //tell WiFi-Class not to connect. Can'T call directly due to async call
			WiFi.disconnect(true);  //disconnect
			Server.end();	
			MDNS.end();
		}
		break;
    case ARDUINO_EVENT_ETH_GOT_IP:
		#if (DBG_ON && DBG_WiFi)
			Serial.print("Ethernet:\r\n MAC: "); Serial.print(ETH.macAddress());
			Serial.print("\r\n IPv4: ");Serial.print(ETH.localIP());
			if (ETH.fullDuplex()) 
				Serial.print("\r\n FULL_DUPLEX");
			Serial.print("\r\n Speed"); Serial.print(ETH.linkSpeed()); Serial.println("Mbps\r\n");
		#endif
		if(ETH.localIP() !=0)
		{
			WiFi_EA.setEthernet_connected(true);
			MDNS.begin(ETHERNETHOSTNAME); //mDNS is used for beautified name broadcasting in networks.
			#if (DBG_ON && DBG_WiFi)
				Serial.println("Ethernet Server begin");
			#endif
			Server.begin();
		}
		else
			WiFi_EA.setWifi_reconnect_needed(true); //tell WiFi-Class to connect. Can't call directly due to async call
		break;
    case ARDUINO_EVENT_ETH_DISCONNECTED:
		#if (DBG_ON && DBG_WiFi)
			Serial.println("Ethernet: Disconnected");
			Serial.printf("WiFi connect: %d to %s PW: %s\r\n", WiFi_EA.ethernet_connected(), WiFi_EA.userSSID(), WiFi_EA.userPassword());
		#endif
		if(WiFi_EA.ethernet_connected())
		{
			WiFi_EA.setEthernet_connected(false); //Wired lan disconnect
			WiFi.disconnect(true); //Set all network classes to disconnected
			Server.end();	
			MDNS.end();
			WiFi_EA.setWifi_reconnect_needed(true); //tell WiFi-Class to connect. Can't call directly due to async call
		}
		break;
    case ARDUINO_EVENT_ETH_STOP:
		#if (DBG_ON && DBG_WiFi)
			Serial.println("Ethernet: Stopped");
		#endif
		if(WiFi_EA.ethernet_connected()) //If ethernet is still connected but crashed somewhere switch it off in this event
		{
			WiFi_EA.setEthernet_connected(false);
			WiFi.disconnect(true); 
			Server.end();	
			MDNS.end();
			WiFi_EA.setWifi_reconnect_needed(true);
		}
		break;
	case ARDUINO_EVENT_WIFI_STA_DISCONNECTED:
		#if (DBG_ON && DBG_WiFi)
			Serial.println("Disconnected from WIFI access point");
			Serial.print("WiFi lost connection.");
			Serial.println("Reconnecting...");
		#endif
		WiFi_EA.setWifi_reconnect_needed(true); //tell WiFi-Class to connect. Can't call directly due to async call
		break;
    default:
		#if (DBG_ON && DBG_WiFi)
			Serial.printf("WiFi event: %d\r\n", event);
		#endif
		break;
  }
}

/*----------------------------
	Func: setup
	Desc: Initialize ethernet (wird lan LAN8720)
	Vars: ---
	Ret : ---
------------------------------*/
void wifi_ea::setup()
{
	eth_connected = false; //Variables for internal use in class
	reconnect_needed = false;
	#if (DBG_ON && DBG_WiFi)
		Serial.println("Ethernet on");
	#endif
	pinMode(23, OUTPUT); //Activate clock to LAN8720-IC
	digitalWrite(23,HIGH);
	delay(5); //Wait for stable clock
	WiFi.onEvent(WiFiEvent); //connect WiFi / Ethernet event
	ETH.begin(0, -1, 16, 17, ETH_PHY_LAN8720); //Ethernet on, define pins and IC
	#if (DBG_ON && DBG_WiFi)
		Serial.println("Ethernet ready");
	#endif
	WiFi.setSleep(false);	
}

/*----------------------------
	Func: connectWiFi
	Desc: Connects to WiFi
	Vars: ssid, password
	Ret : ---
------------------------------*/
void wifi_ea::connectWiFi(char *ssid, char* password)
{
	strcpy(this->ssid, ssid); //remember connection for function reconnect()
	strcpy(pw, password);
	
	#if (DBG_ON && DBG_WiFi)
		Serial.printf("WLAN to %s PW: %s. Ethernet connected? %d\r\n", ssid, password, eth_connected);
	#endif	
	if(eth_connected) //IF ethernet already connected, do not connect WiFi
		return;
	int status = WL_DISCONNECTED;
	if(WiFi.status() == WL_DISCONNECTED || strcmp(WiFi.SSID().c_str(),ssid)!=0) //already connected to another wifi or unconnected?
	{
		if(WiFi.status() == WL_CONNECTED) //already connected to another wifi
		{
			WiFi.disconnect(true);  //disconnect
			Server.end();	
			MDNS.end();
		}
		WiFi.hostname(HOSTNAME); 
		#if (DBG_ON && DBG_WiFi)
			Serial.printf("Connect WLAN to %s PW: %s\r\n", ssid, password, eth_connected);
		#endif	
		WiFi.begin (ssid, password); //connect to SSID with PW

		status = WiFi.waitForConnectResult(CONNECTION_TIMEOUT);
		/*unsigned long myTime = millis()+(CONNECTION_TIMEOUT*10);	
		while (WiFi.status() != WL_CONNECTED && myTime>=millis()) //wait for stable connection
		;*/
	}
	
	#if (DBG_ON && DBG_WiFi)
		Serial.printf("Connection to %s Status: %d\r\n", ssid, WiFi.status());
		Serial.printf("Connection to Lan Status: %d, %d.%d.%d.%d\r\n", ETH.linkUp(), ETH.localIP()[0],ETH.localIP()[1],ETH.localIP()[2],ETH.localIP()[3]);
	#endif	
	
	if(status == WL_CONNECTED ) //connection successful?
	{
		reconnect_needed = false;
		MDNS.begin(HOSTNAME); //mDNS is used for beautified name broadcasting in networks.
		#if (DBG_ON && DBG_WiFi)
			Serial.println("WiFi Server begin");
		#endif
		Server.begin();
	}
	else //connection denied
	{
		if(status != WL_CONNECTION_LOST)
			reconnect_needed = false;
		WiFi.disconnect(); 
		Server.end();	
		MDNS.end();
	}
}

/*----------------------------
	Func: disconnect
	Desc: Disconnects WiFi
	Vars: ---
	Ret : ---
------------------------------*/
void wifi_ea::disconnect()
{
	if(WiFi.status() == WL_CONNECTED)
	{
		WiFi.disconnect(true);  
		Server.end();	
		MDNS.end();
	}
}

/*----------------------------
	Func: reconnect
	Desc: Connects WiFi to previous one
	Vars: ---
	Ret : ---
------------------------------*/
void wifi_ea::reconnect()
{
	connectWiFi(ssid, pw);
}

/*----------------------------
	Func: isConnected
	Desc: Returns true if WiFi is connected
	Vars: ---
	Ret : true on connected
------------------------------*/
boolean wifi_ea::isConnected()
{
	return (WiFi.status() == WL_CONNECTED || ETH.localIP()!= 0);
}

/*----------------------------
	Func: localIP
	Desc: Actual IP-Adress retrieved from DNS server
	Vars: ---
	Ret : byte[4] as IPAddress
------------------------------*/
IPAddress wifi_ea::localIP()
{
	if(eth_connected)
		return ETH.localIP();
	return WiFi.localIP();
}

/*----------------------------
	Func: scanNetworks
	Desc: Count of available networks
	Vars: ---
	Ret : network-count
------------------------------*/
int wifi_ea::scanNetworks()
{
	return WiFi.scanNetworks();
}

/*----------------------------
	Func: SSID
	Desc: Actual connected SSID name
	Vars: ---
	Ret : String with SSID name
------------------------------*/
String wifi_ea::SSID()
{
	if(eth_connected)
		return "Ethernet";
	return WiFi.SSID();
}

/*----------------------------
	Func: SSID
	Desc: Returns SSID name from available net (scanNetworks before calling this function)
	Vars: net number
	Ret : String with SSID name
------------------------------*/
String wifi_ea::SSID(int net)
{
	return WiFi.SSID(net);
}

/*----------------------------
	Func: RSSI
	Desc: Returns actual connected network strength (100 on ethernet)
	Vars: ---
	Ret : strength
------------------------------*/
int wifi_ea::RSSI()
{
	if(eth_connected)
		return 1000;
	return WiFi.RSSI();	
}

/*----------------------------
	Func: RSSI
	Desc: Returns network strength from available net (scanNetworks before calling this function)
	Vars: net number
	Ret : strength
------------------------------*/
int wifi_ea::RSSI(int net)
{
	return WiFi.RSSI(net);
}

/*----------------------------
	Func: SSID
	Desc: Returns SSID name from available net (scanNetworks before calling this function)
	Vars: net number
	Ret : String with SSID name
------------------------------*/
const char* wifi_ea::Hostname()
{
	if(eth_connected)
		return ETH.getHostname();
	return WiFi.getHostname();
}

/*----------------------------
	Func: WiFiServerReceivedData
	Desc: Checks if TCP client is connected. If yes, establish direct connection between TCP client and uniTFT. No protocol buffer in between
	Vars: ---
	Ret : ---
------------------------------*/
void wifi_ea::WiFiServerReceivedData()
{
	if(!isConnected())
		return;
	WiFiRemoteClient = Server.available(); 
	if(WiFiRemoteClient)
	{
		uniTFT.initialize_rs232_newBaud(BAUDUNITFT_DIRECT); //change baudrate to uniTFT (default increase speed to maximum)
		int rec;
		#if (DBG_ON && DBG_WiFi)
			Serial.println("WiFi Client connected");
		#endif
		delay(5);
		while (WiFiRemoteClient.connected()) //as long as client is connected do nothing else
		{
			while(uniTFT.rs232_direct_available() && WiFiRemoteClient.connected()) //Data from uniTFT received -> send to WiFi client
			{
				rec = uniTFT.rs232_direct_read(WiFiServerReceiveBuffer, sizeof(WiFiServerReceiveBuffer));
				WiFiRemoteClient.write(WiFiServerReceiveBuffer, rec);
			}

			while(WiFiRemoteClient.connected() && WiFiRemoteClient.available())  //Data from WiFi client received -> send to uniTFT
			{
				rec = WiFiRemoteClient.read(WiFiServerReceiveBuffer, sizeof(WiFiServerReceiveBuffer));	
				char *p_rec_buf = (char*)WiFiServerReceiveBuffer;
				while(rec--) 
					uniTFT.rs232_direct_write(*p_rec_buf++);
			}
		}
		#if (DBG_ON && DBG_WiFi)
			Serial.println("WiFi Client disconnected");
		#endif
		uniTFT.initialize_rs232_newBaud(BAUDUNITFT); //set baudrate back to default after disconnect from client
		#if (DBG_ON && DBG_WiFi)
			Serial.printf("Back to %d\r\n", BAUDUNITFT);
		#endif

		delay(10);
	}
}